#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, kernels

#include "gpu_blastp.h"
#include "gpu_blastp_kernel.cu"
#include <algo/blast/core/gpu_cpu_common.h>
#include "gpu_blastp_kernel.h"

#include <algo/blast/core/blast_aalookup.h>

#include <algorithm>
#include <vector>

using namespace std;

#define CUDA_DEVICE_MAXNUMBER 100
/// The macro "NO_CELL_LEVELS" shows the statistic
/// function will report at most 2^(NO_CELL_LEVELS-1)
/// results can be stored within a subject.
#define NO_CELL_LEVELS 7


typedef struct
{
    int GPU_device_id;
    char* h_db_access_entry; //database first ptr
    int* h_db_Sequence_Maxlength_vector; //the vector addressed max length of each group of subjects
    Int4 number_of_Groups; // the size of vector @h_db_Sequence_Maxlength_vector
    Int4 num_seq_to; // the no. of sequence in a part of one file volumne
    unsigned long h_GPU_BLASTP_database_bytes; // the total bytes of the related database section
    Int4 h_Hits_offset; // The first address of the location in the space @h_Hits
    Int4 h_GPUBlastInitHitList_offset; //The first address of the location in the space @h_GPUBlastInitHitList
    Int4 h_Hits_bytes;
    Int4 h_GPUBlastInitHitList_bytes;

    Int4* h_Hits;
    GPUBlastInitHitList* h_GPUBlastInitHitList;

    Int4* d_Hits;
    GPUBlastInitHitList* d_GPUBlastInitHitList;


    Int4* d_Database2Dpadded;
    Int4* d_RepeatedSubstitutionMatrix;
    Int4* d_RepeatedSequence_length_vector;
    PV_ARRAY_TYPE* d_RepeatedPV;
    AaLookupSmallboneCell* d_ThickBackbone;
    Uint2* d_overflow;
    short* d_RepeatedDiag_array;

    cudaStream_t stream_id;
    cudaEvent_t event_id, start;

} GPU_Stream_Block;

static unsigned int MAX_SUCCESSFUL_HITS_PER_SEQUENCE=19;


/// For query adjustment
void static small_bobbo_sorting(unsigned short* short_array, int array_count )
{
    int i,j;
    Uint2 tmp=0;
    for(i=0; i<array_count; i++)
        for(j=i+1; j<array_count; j++)
        {
            if(short_array[j]<short_array[j-1])
            {
                tmp = short_array[j];
                short_array[j] = short_array[j-1];
                short_array[j-1] = tmp;
            }
        }
    return ;
}


void lut_sorting(BlastAaLookupTable *lookup)
{


    AaLookupSmallboneCell* h_ThickBackbone = (AaLookupSmallboneCell*) lookup->thick_backbone;

    unsigned short* h_overflow = (unsigned short*)lookup->overflow;

    unsigned long int h_overflow_count = lookup->overflow_size;

    vector<unsigned short> overflow_vector (h_overflow, h_overflow+h_overflow_count);

    int i;

    for(i=0; i<lookup->backbone_size; i++)
    {
        if(h_ThickBackbone[i].num_used > AA_HITS_PER_CELL)
            sort(overflow_vector.begin()+ h_ThickBackbone[i].payload.overflow_cursor,  overflow_vector.begin() + h_ThickBackbone[i].payload.overflow_cursor + h_ThickBackbone[i].num_used  );
        else if(h_ThickBackbone[i].num_used > 0 )
            small_bobbo_sorting( h_ThickBackbone[i].payload.entries,  h_ThickBackbone[i].num_used);
    }

    for(i=0; i<h_overflow_count ; i++)
    {
        h_overflow[i] = overflow_vector[i];
    }

    return;

}


/// For adjustment the ratio used in GPU DB ==============
static GPU_Stream_Block stream_blocks[CUDA_DEVICE_MAXNUMBER];
static double current_gpu_walltime[CUDA_DEVICE_MAXNUMBER];
static double max_gpu_walltime;

double get_gpu_walltime();

void set_max_gpu_walltime();

void init_max_and_current_gpu_walltime();

int get_num_of_gpu_cards_used();

static int num_of_gpu_cards_used;

Int4* local_h_Hits=NULL;
GPUBlastInitHitList* local_h_GPUBlastInitHitList=NULL;

char* h_Database2Dpadded=NULL;

void gpu_set_device(const int device_id);
void gpu_get_device_count(int * count);

double get_gpu_walltime()
{
    return max_gpu_walltime;
}

int get_num_of_gpu_cards_used()
{
    return num_of_gpu_cards_used;
}
void set_max_gpu_walltime()
{
    int i;
    double max_record=0;
    for(i=0; i<num_of_gpu_cards_used; i++)
    {
        if(max_record<current_gpu_walltime[i])
            max_record = current_gpu_walltime[i];
    }
    max_gpu_walltime= max_record;
}

void init_max_and_current_gpu_walltime()
{
    int i;
    for(i=0; i<num_of_gpu_cards_used; i++)
        current_gpu_walltime[i]=0;

    max_gpu_walltime=0;
}

void set_MAX_SUCCESSFUL_HITS_PER_SEQUENCE(const unsigned int val)
{
    MAX_SUCCESSFUL_HITS_PER_SEQUENCE = val;
}

unsigned int get_MAX_SUCCESSFUL_HITS_PER_SEQUENCE()
{
    return MAX_SUCCESSFUL_HITS_PER_SEQUENCE;
}

void SetNumOfGpuCards(const int expected_num_of_cards)
{
    gpu_get_device_count(&num_of_gpu_cards_used);
    if(num_of_gpu_cards_used>CUDA_DEVICE_MAXNUMBER)
    {
        printf("Warning: The no. of avalible gpu cards are more then %d.\nNow, uses %d cards only!\nTo use all cards, please modify the constant CUDA_DEVICE_MAXNUMBER\n",
               CUDA_DEVICE_MAXNUMBER, num_of_gpu_cards_used);
        num_of_gpu_cards_used = CUDA_DEVICE_MAXNUMBER;
    }
    if( (num_of_gpu_cards_used>=expected_num_of_cards) && (0!=expected_num_of_cards))
        num_of_gpu_cards_used = expected_num_of_cards;
    else if(num_of_gpu_cards_used < expected_num_of_cards)
    {
        printf("Warning: There is not enough gpu cards avalible, uses %d cards only!\n", num_of_gpu_cards_used);
    }

}

void gpu_quit()
{
    /// !!!!!!!!!!!!!!!!!!!!
    /// Note that this function will reset the device immediately.
    /// It is the caller's responsibility to ensure that the device
    /// is not being accessed by any other host threads from the
    /// process when this function is called.
    /// !!!!!!!!!!!!!!!!!!!!
    cudaDeviceReset();
}


void gpu_abnormally_exit(int exit_code)
{
    /// !!!!!!!!!!!!!!!!!!!!
    /// Note that this function will reset the device immediately.
    /// It is the caller's responsibility to ensure that the device
    /// is not being accessed by any other host threads from the
    /// process when this function is called.
    /// !!!!!!!!!!!!!!!!!!!!
    gpu_quit();
    exit(exit_code);
}

void all_gpu_quit()
{
    /// !!!!!!!!!!!!!!!!!!!!
    /// Note that this function will reset the device immediately.
    /// It is the caller's responsibility to ensure that the device
    /// is not being accessed by any other host threads from the
    /// process when this function is called.
    /// !!!!!!!!!!!!!!!!!!!!
    int gpu_device_count=num_of_gpu_cards_used, i;

    for(i=0; i<gpu_device_count; i++)
    {
        gpu_set_device(i);
        cudaDeviceReset();
    }

}

void all_gpu_abnormally_exit(int exit_code)
{
    all_gpu_quit();
    exit(exit_code);
}


void gpu_set_cache_config(enum cudaFuncCache cacheConfig)
{
    cudaError_t return_code = cudaFuncSetCacheConfig(GPU_BLASTP_kernelTwoHit, cacheConfig);
    if(cudaSuccess  != return_code)
    {
        fprintf(stderr, "cudaFuncSetCacheConfig error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}

void gpu_memcpy_async(void * dst,
                      const void *  src,
                      size_t  count,
                      enum cudaMemcpyKind  kind,
                      cudaStream_t stream)
{
    cudaError_t return_code = cudaMemcpyAsync (dst, src, count, kind, stream);
    if(cudaSuccess  != return_code)
    {
        fprintf(stderr, "cudaMemcpyAsync error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}


void gpu_memcpy(void * dst,
                const void *  src,
                size_t  count,
                enum cudaMemcpyKind  kind
               )
{
    cudaError_t return_code = cudaMemcpy(dst, src, count, kind);
    if(cudaSuccess  != return_code)
    {
        fprintf(stderr, "cudaMemcpy error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}



void gpu_set_device(const int device_id)
{
    cudaError_t return_code = cudaSetDevice(device_id);
    if(cudaSuccess  != return_code)
    {

        fprintf(stderr, "cudaSetDevice error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;

}

void gpu_device_mem_free(void ** d_ptr)
{
    if(NULL==*d_ptr)
    {
        fprintf(stderr, "gpu_device_mem_free error: input pointer is NULL\n");
        all_gpu_abnormally_exit(-1);
    }
    cudaError_t return_code = cudaFree(*d_ptr);
    if(cudaSuccess != return_code)
    {

        fprintf(stderr, "cudaFree error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    *d_ptr = NULL;
    return;
}



void gpu_host_calloc(void ** ptr,
                     size_t  size,
                     unsigned int  flags=cudaHostAllocDefault
                    )
{
    cudaError_t return_code = cudaHostAlloc(ptr, size, flags);
    if(cudaSuccess != return_code)
    {
        fprintf(stderr, "cudaHostAlloc error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    memset(*ptr, 0, size);
    return;

}


void gpu_device_calloc(void ** d_ptr, size_t size)
{
    cudaError_t return_code = cudaMalloc(d_ptr, size);
    if(cudaSuccess != return_code)
    {
        fprintf(stderr, "cudaMalloc error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    cudaMemset(*d_ptr, 0, size);
    return;

}


void gpu_host_mem_free(void** h_ptr)
{

    if(NULL==*h_ptr)
    {
        fprintf(stderr, "gpu_host_mem_free error: input pointer is NULL\n");
        all_gpu_abnormally_exit(-1);
    }
    cudaError_t return_code = cudaFreeHost(*h_ptr);
    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaFreeHost error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    *h_ptr = NULL;
    return ;
}


void gpu_memcpy_to_symbol(const char *  symbol,
                          const void *  src,
                          size_t  count,
                          size_t  offset=0,
                          enum cudaMemcpyKind kind=cudaMemcpyHostToDevice)
{
    cudaError_t return_code = cudaMemcpyToSymbol(symbol, src, count, offset,kind);
    if(cudaSuccess  != return_code)
    {

        fprintf(stderr, "cudaMemcpyToSymbol error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}


void gpu_get_device_count(int * count)
{
    cudaError_t return_code = cudaGetDeviceCount(count);
    if(cudaSuccess  != return_code)
    {

        fprintf(stderr, "cudaGetDeviceCount error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}


void init_stream_obj(int cuda_device_no)
{
    gpu_set_device(cuda_device_no);

    cudaError_t return_code=cudaStreamCreate(&stream_blocks[cuda_device_no].stream_id);

    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaStreamCreate error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }

    return_code=cudaEventCreate(&stream_blocks[cuda_device_no].event_id);

    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaEventCreate error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }

    return_code=cudaEventCreate(&stream_blocks[cuda_device_no].start);

    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaEventCreate error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }

    return;
}

void GPU_Host_Alloc_WC(void** h_ptr, const size_t h_size)
{
    cudaError_t return_code=cudaHostAlloc(h_ptr,  h_size, cudaHostAllocWriteCombined);

    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaHostAlloc error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }
    return;
}


void gpu_stream_synchronize(int cuda_device_id)
{

    cudaError_t return_code;

    gpu_set_device(cuda_device_id);

    return_code=cudaEventSynchronize(stream_blocks[cuda_device_id].event_id);
    if( cudaSuccess !=return_code)
    {
        fprintf(stderr, "cudaEventSynchronize error: %s\n", cudaGetErrorString(return_code));
        all_gpu_abnormally_exit(-1);
    }

    ///=============== Timing report ===============
    float kernel_time;
    cudaEventElapsedTime(&kernel_time, stream_blocks[cuda_device_id].start, stream_blocks[cuda_device_id].event_id);

    printf("Kernel elapses %lg sec\n", kernel_time/1000);

    current_gpu_walltime[cuda_device_id] = kernel_time/1000;

    ///=============== Timing report ===============
    return;
}


/** Read the database used by GPU_BLASTP
 * @param h_GPU_BLASTP_database        array that holds the GPU database [out]
 * @param h_GPU_BLASTP_database_bytes  size in bytes of the GPU_database [out]
*/

void ReadGPU_BLASTP_database(
    const char* GPU_volume_name,
    const unsigned long h_GPU_BLASTP_database_bytes)
{

    FILE * pFile;
    long lSize;
    size_t result;

    pFile = fopen ( GPU_volume_name, "rb" );

    if (pFile==NULL)
    {
        fputs ("File error: GPU database file could not be oppened",stderr);
        all_gpu_abnormally_exit (1);
    }

    lSize = h_GPU_BLASTP_database_bytes;

    if(h_Database2Dpadded!=NULL)
        gpu_host_mem_free((void**)&h_Database2Dpadded);

    gpu_host_calloc((void**)&h_Database2Dpadded, lSize, cudaHostAllocPortable);

    result = fread (h_Database2Dpadded, sizeof(char), lSize, pFile);

    fclose (pFile);

}


void setup_data_partitions(Int4 Group_number,
                           int * Sequence_length_vector,
                           char * h_GPU_BLASTP_database,
                           int cuda_device_count,
                           const Int4 Group_size,
                           const Int4 num_seq_to,
                           const Int4 total_sequence_bytes,
                           const Int4 total_init_hsp_list_length)
{

    Int4 average_sequence_bytes=0, sequence_bytes_of_a_part=0,  group_ptr=0, last_group_ptr=0;
    Int4 last_total_database_bytes=0;
    Int4 i=0;

    if(cuda_device_count==0)
    {
        printf("Error: the cuda device count is 0!\n");
        exit(-1);
    }

    average_sequence_bytes=total_sequence_bytes*1.1/cuda_device_count;
    sequence_bytes_of_a_part = 0;

    last_group_ptr=-1;
    group_ptr=-1;
    last_total_database_bytes=0;

    if (1==cuda_device_count)
    {
        stream_blocks[0].h_db_Sequence_Maxlength_vector=Sequence_length_vector;
        stream_blocks[0].number_of_Groups=Group_number;
        stream_blocks[0].h_GPU_BLASTP_database_bytes = total_sequence_bytes ;
        stream_blocks[0].h_Hits_offset =0;
        stream_blocks[0].h_GPUBlastInitHitList_offset = 0;
        stream_blocks[0].num_seq_to = num_seq_to ;
        stream_blocks[0].h_db_access_entry= h_GPU_BLASTP_database;
    }
    else
    {

        char flag=0;

        for(i=0; i< cuda_device_count-1; i++)
        {
            stream_blocks[i].h_db_access_entry= &(h_GPU_BLASTP_database[last_total_database_bytes]);

            while(group_ptr +1< Group_number-1)
            {
                sequence_bytes_of_a_part +=  Sequence_length_vector[group_ptr+1]*Group_size;

                if(sequence_bytes_of_a_part>average_sequence_bytes)
                {
                    stream_blocks[i].h_db_Sequence_Maxlength_vector=&(Sequence_length_vector[last_group_ptr+1]);
                    stream_blocks[i].number_of_Groups=group_ptr-last_group_ptr;
                    stream_blocks[i].num_seq_to = stream_blocks[i].number_of_Groups*Group_size;
                    if(group_ptr==last_group_ptr)
                    {
                        printf("Error: The size of a group of subject sequences is greater than the amount of average sequence bytes of a part! \
                               Please decrease the gpu thread size or the gpu block size!\n");
                        exit(-1);
                    }
                    else
                    {
                        stream_blocks[i].h_GPU_BLASTP_database_bytes = sequence_bytes_of_a_part - Sequence_length_vector[group_ptr+1]*Group_size;
                    }

                    stream_blocks[i].h_Hits_offset = (last_group_ptr+1)*Group_size*H_HITS_SIZE_BYTE;
                    stream_blocks[i].h_GPUBlastInitHitList_offset = i*total_init_hsp_list_length/cuda_device_count;

                    last_total_database_bytes += stream_blocks[i].h_GPU_BLASTP_database_bytes;
                    sequence_bytes_of_a_part = 0;

                    last_group_ptr=group_ptr;
                    break;
                }
                group_ptr ++;
            }

            if(group_ptr+1 == Group_number-1)
            {
                if(0==flag)
                {
                    stream_blocks[i].h_GPU_BLASTP_database_bytes = total_sequence_bytes - last_total_database_bytes;

                    if((group_ptr==last_group_ptr) && (stream_blocks[i].h_GPU_BLASTP_database_bytes > average_sequence_bytes))
                    {
                        printf("Error: The size of the last group of subject sequences is greater than the amount of average sequence bytes of a part! \
                               Please decrease the gpu thread size or the gpu block size!\n");
                        exit(-1);
                    }

                    else if(stream_blocks[i].h_GPU_BLASTP_database_bytes > average_sequence_bytes)
                    {
                        flag = 1;
                        stream_blocks[i].h_GPU_BLASTP_database_bytes = sequence_bytes_of_a_part;
                        stream_blocks[i].h_db_Sequence_Maxlength_vector=&(Sequence_length_vector[last_group_ptr+1]);
                        stream_blocks[i].number_of_Groups=group_ptr-last_group_ptr;
                        stream_blocks[i].h_Hits_offset = (last_group_ptr+1)*Group_size*H_HITS_SIZE_BYTE;
                        stream_blocks[i].h_GPUBlastInitHitList_offset = i*total_init_hsp_list_length/cuda_device_count;
                        stream_blocks[i].num_seq_to = stream_blocks[i].number_of_Groups*Group_size;
                        last_group_ptr=group_ptr;
                        last_total_database_bytes += stream_blocks[i].h_GPU_BLASTP_database_bytes;
                        sequence_bytes_of_a_part=0;
                    }

                    else
                    {

                        stream_blocks[i].h_db_Sequence_Maxlength_vector=&(Sequence_length_vector[Group_number-1]);
                        stream_blocks[i].number_of_Groups=1;

                        stream_blocks[i].h_Hits_offset = (last_group_ptr+1)*Group_size*H_HITS_SIZE_BYTE;
                        stream_blocks[i].h_GPUBlastInitHitList_offset =  i*total_init_hsp_list_length/cuda_device_count;
                        stream_blocks[i].num_seq_to = num_seq_to - (last_group_ptr+1)*Group_size;

                        flag = 2;
                        group_ptr++;
                        last_total_database_bytes = total_sequence_bytes;
                        last_group_ptr = Group_number-1;
                    }
                }
                else if(1==flag)
                {

                    stream_blocks[i].h_db_Sequence_Maxlength_vector=&(Sequence_length_vector[last_group_ptr+1]);
                    stream_blocks[i].number_of_Groups=1;
                    stream_blocks[i].h_GPU_BLASTP_database_bytes = total_sequence_bytes - last_total_database_bytes;
                    if(stream_blocks[i].h_GPU_BLASTP_database_bytes > average_sequence_bytes)
                    {
                        printf("Error: The size of the last group of subject sequences is greater than the \n amount of average sequence bytes of a part! \n \
                               Please decrease the gpu thread size or the gpu block size!\n");
                        exit(-1);
                    }
                    stream_blocks[i].h_Hits_offset = (Group_number-1)*Group_size*H_HITS_SIZE_BYTE;
                    stream_blocks[i].h_GPUBlastInitHitList_offset = i*total_init_hsp_list_length/cuda_device_count;

                    stream_blocks[i].num_seq_to = num_seq_to - (Group_number-1)*Group_size;

                    flag=2;
                    group_ptr++;
                    last_total_database_bytes = total_sequence_bytes;
                    last_group_ptr = Group_number-1;
                }

            }

            if(group_ptr+1 == Group_number)
            {
                stream_blocks[i].h_db_Sequence_Maxlength_vector=Sequence_length_vector;
                stream_blocks[i].number_of_Groups=0;
                stream_blocks[i].h_GPU_BLASTP_database_bytes = 0;
                stream_blocks[i].h_Hits_offset = 0;
                stream_blocks[i].h_GPUBlastInitHitList_offset = 0;
                stream_blocks[i].num_seq_to = 0;
            }

        }

        if(group_ptr+1 == Group_number)
        {
            stream_blocks[cuda_device_count-1].h_db_access_entry= h_GPU_BLASTP_database;
            stream_blocks[cuda_device_count-1].h_db_Sequence_Maxlength_vector=Sequence_length_vector;
            stream_blocks[cuda_device_count-1].number_of_Groups=0;
            stream_blocks[cuda_device_count-1].h_GPU_BLASTP_database_bytes = 0;
            stream_blocks[cuda_device_count-1].h_Hits_offset = 0;
            stream_blocks[cuda_device_count-1].h_GPUBlastInitHitList_offset = 0;
            stream_blocks[cuda_device_count-1].num_seq_to = 0;
        }
        else
        {
            stream_blocks[cuda_device_count-1].h_db_access_entry= &(h_GPU_BLASTP_database[last_total_database_bytes]);
            stream_blocks[cuda_device_count-1].h_db_Sequence_Maxlength_vector=&(Sequence_length_vector[last_group_ptr+1]);
            stream_blocks[cuda_device_count-1].number_of_Groups=Group_number-last_group_ptr-1;
            stream_blocks[cuda_device_count-1].h_GPU_BLASTP_database_bytes = total_sequence_bytes - last_total_database_bytes;


            if((group_ptr==last_group_ptr) && (stream_blocks[cuda_device_count-1].h_GPU_BLASTP_database_bytes > average_sequence_bytes))
            {
                printf("Warning: The size of the last group of subject sequences is greater than the amount of average sequence bytes of a part! \
                               Please decrease the gpu thread size or the gpu block size!\n");
            }
            stream_blocks[cuda_device_count-1].h_Hits_offset = (last_group_ptr+1)*Group_size*H_HITS_SIZE_BYTE;
            stream_blocks[cuda_device_count-1].h_GPUBlastInitHitList_offset = (cuda_device_count-1)*total_init_hsp_list_length/cuda_device_count;
            stream_blocks[cuda_device_count-1].num_seq_to = num_seq_to - (last_group_ptr+1)*Group_size;
        }
    }

    for(i=0; i<cuda_device_count; i++)
    {
        stream_blocks[i].h_Hits_bytes =   stream_blocks[i].num_seq_to * H_HITS_SIZE_BYTE * sizeof(Int4);
        stream_blocks[i].h_GPUBlastInitHitList_bytes = total_init_hsp_list_length * sizeof(GPUBlastInitHitList) / cuda_device_count;
        stream_blocks[i].h_Hits = NULL;
        stream_blocks[i].h_GPUBlastInitHitList = NULL;
    }

    for(i=0; i<cuda_device_count; i++)
    {
        if(stream_blocks[i].num_seq_to==0) stream_blocks[i].GPU_device_id = -1;
    }
    return;
}


void  GPU_BLASTP_pre(
    const BLAST_SequenceBlk* query, const BlastQueryInfo* query_info,
    const LookupTableWrap* lookup_wrap,
    const Blast_ExtendWord* ewp, const BlastInitialWordParameters* word_params,
    const BlastGPUOptions* gpu_options, Int4* Sequence_length_vector,
    const int h_Database2Dpadded_bytes,
    const Int4 Group_number,
    const Int4 h_Hits_bytes, const Int4 h_GPUBlastInitHitList_bytes
)
{

    if(local_h_Hits!=NULL) gpu_host_mem_free((void**)&local_h_Hits);
    if(local_h_GPUBlastInitHitList!=NULL) gpu_host_mem_free((void**)&local_h_GPUBlastInitHitList);

    gpu_host_calloc((void**)&local_h_Hits, h_Hits_bytes, cudaHostAllocPortable);
    gpu_host_calloc((void**)&local_h_GPUBlastInitHitList, h_GPUBlastInitHitList_bytes, cudaHostAllocPortable);

    int gpu_device_count=num_of_gpu_cards_used;


    setup_data_partitions(Group_number,
                          Sequence_length_vector,
                          h_Database2Dpadded,
                          gpu_device_count,
                          gpu_options->num_blocksx * gpu_options->num_threadsx,
                          gpu_options->num_sequences_to,
                          h_Database2Dpadded_bytes,
                          h_GPUBlastInitHitList_bytes/sizeof(GPUBlastInitHitList));

    for(int cuda_device_id=0; cuda_device_id<gpu_device_count; cuda_device_id++)
    {
        init_stream_obj(cuda_device_id);

        GPU_BLASTP(cuda_device_id,
                   query, query_info,
                   lookup_wrap,
                   ewp, word_params,
                   gpu_options,
                   stream_blocks[cuda_device_id].h_db_Sequence_Maxlength_vector
                  );
    }
}

void  GPU_BLASTP(const int cuda_device_id,
    const BLAST_SequenceBlk* query, const BlastQueryInfo* query_info,
    const LookupTableWrap* lookup_wrap,
    const Blast_ExtendWord* ewp, const BlastInitialWordParameters* word_params,
    const BlastGPUOptions* gpu_options, const Int4* Sequence_length_vector
)
{

    const Int4 num_threadsy = 1, num_blocksy = 1;
    const Int4 num_blocksx = gpu_options->num_blocksx, num_threadsx = gpu_options->num_threadsx,  num_sequences = stream_blocks[cuda_device_id].num_seq_to;//gpu_options->num_sequences_to;

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Number of processed sequences by the GPU = %5d\n", num_sequences);

    int Group_number = (num_sequences / (num_threadsx*num_blocksx) ) + (( (num_sequences % (num_threadsx*num_blocksx) ) == 0)?0:1);

    int Sequence_length_vector_stride = ( Group_number / PITCH + ((Group_number % PITCH)?1:0) ) * PITCH;
    int h_RepeatedSequence_length_vector_integers = Sequence_length_vector_stride * num_blocksx;
    unsigned long int h_RepeatedSequence_length_vector_bytes = Sequence_length_vector_stride * num_blocksx * sizeof(int);

    int* h_RepeatedSequence_length_vector=(int*) calloc( h_RepeatedSequence_length_vector_integers, sizeof(int)  );

    MultipleCopyMaxlength(h_RepeatedSequence_length_vector, Sequence_length_vector, Group_number, Sequence_length_vector_stride, num_blocksx);

    BlastAaLookupTable *lookup = (BlastAaLookupTable*) lookup_wrap->lut;

    PV_ARRAY_TYPE *h_pv = lookup->pv;
    int pv_length = (lookup->backbone_size >> PV_ARRAY_BTS) + 1;

    Round2Multiple(&pv_length);

    unsigned long int h_RepeatedPV_bytes = num_blocksx * pv_length * sizeof(PV_ARRAY_TYPE);

    PV_ARRAY_TYPE* h_RepeatedPV = (PV_ARRAY_TYPE*) calloc( h_RepeatedPV_bytes, sizeof(char));


    MultipleCopyPV(h_RepeatedPV, h_pv, pv_length, num_blocksx);

    AaLookupSmallboneCell* h_ThickBackbone = (AaLookupSmallboneCell*) lookup->thick_backbone;

    Uint2* h_overflow = (Uint2*)lookup->overflow;

    unsigned long int h_ThickBackbone_bytes = (lookup->backbone_size+1) * sizeof(AaLookupSmallboneCell);
    unsigned long int h_overflow_bytes = (lookup->overflow_size+1) * sizeof(Uint2);

    unsigned int diag_array_length = ewp->diag_table->diag_array_length;

    unsigned long long int h_RepeatedDiag_array_bytes = (unsigned long long int)diag_array_length * num_blocksx*num_threadsx*sizeof(short);
    int diag_mask = ewp->diag_table->diag_mask;
    int window = ewp->diag_table->window;


    int Query_length = query->length;
    unsigned char* h_Query = query->sequence;

    int num_contexts_invoked = query_info->last_context - query_info->first_context +1;

    int* h_x_dropoff = (int*)malloc(num_contexts_invoked * sizeof(int) );
    int* h_cutoff_score = (int*)malloc(num_contexts_invoked * sizeof(int) );

    int* h_query_offset = (int*)malloc((num_contexts_invoked+1) * sizeof(int) );

    for(int i = 0; i < num_contexts_invoked; ++i)
    {
        h_x_dropoff[i] = word_params->cutoffs[i].x_dropoff;
        h_cutoff_score[i] = word_params->cutoffs[i].cutoff_score;
        h_query_offset[i] = query_info->contexts[i].query_offset;
    }
    h_query_offset[num_contexts_invoked] = query->length;

    //Read Substitution Matrix
    char* h_SubstitutionMatrix = (char*) calloc( SUBSTITUTION_MATRIX_LENGTH * sizeof(char), sizeof(char) );

    unsigned int h_RepeatedSubstitutionMatrix_bytes = SUBSTITUTION_MATRIX_LENGTH * num_blocksx * sizeof(char);
    char* h_RepeatedSubstitutionMatrix = (char*) calloc ( h_RepeatedSubstitutionMatrix_bytes, sizeof(char) );


    ReadSubstitutionMatrix( h_SubstitutionMatrix, h_RepeatedSubstitutionMatrix, SUBSTITUTION_MATRIX_LENGTH, num_blocksx );

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Global memory allocated memory size = %lf Mb\n",
                (double)(stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes +
                        num_blocksx * num_threadsx * sizeof(int) +
                         h_RepeatedSubstitutionMatrix_bytes +
                         h_RepeatedSequence_length_vector_bytes +
                         stream_blocks[cuda_device_id].h_Hits_bytes +
                         h_RepeatedPV_bytes+
                         h_ThickBackbone_bytes+
                         h_overflow_bytes+
                         h_RepeatedDiag_array_bytes +
                         stream_blocks[cuda_device_id].h_GPUBlastInitHitList_bytes)/1e6);

    unsigned long long int BytesTransferredToGlobalMemory =
        stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes + 2* num_blocksx * num_threadsx * sizeof(int) +
        h_RepeatedSubstitutionMatrix_bytes +
        h_RepeatedSequence_length_vector_bytes +
        h_RepeatedPV_bytes+
        h_ThickBackbone_bytes+
        h_overflow_bytes;

    if( GPU_VERBOSE )
    {
        fprintf(stderr,"\nthreadId = 0: Bytes transferred to global memory\n");
        fprintf(stderr,"\n\nthreadId = 0: Bytes allocated on host\n");
        fprintf(stderr,"threadId = 0: Database           = %d bytes\n",stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes );
        fprintf(stderr,"threadId = 0: SubstitutionMatrix = %d bytes\n",h_RepeatedSubstitutionMatrix_bytes );
        fprintf(stderr,"threadId = 0: SequenceMaxlength  = %d bytes\n",h_RepeatedSequence_length_vector_bytes );
        fprintf(stderr,"threadId = 0: Hits               = %d bytes\n",stream_blocks[cuda_device_id].h_Hits_bytes );
        fprintf(stderr,"threadId = 0: PV                 = %d bytes\n",h_RepeatedPV_bytes );
        fprintf(stderr,"threadId = 0: ThickBackbone      = %d bytes\n",h_ThickBackbone_bytes );
        fprintf(stderr,"threadId = 0: Overflow           = %d bytes\n",h_overflow_bytes );
        fprintf(stderr,"threadId = 0: RepeatedDiagArray  = %d bytes\n",h_RepeatedDiag_array_bytes) ;
        fprintf(stderr,"threadId = 0: InitHitList        = %d bytes\n",stream_blocks[cuda_device_id].h_GPUBlastInitHitList_bytes ) ;

        fprintf(stderr,"\nthreadId = 0: Bytes transferred to global memory\n");
        fprintf(stderr,"threadId = 0: Database           = %10.1f Mb\n",(float)stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes / 1e6);
        fprintf(stderr,"threadId = 0: SubstitutionMatrix = %10.1f Mb\n",(float)h_RepeatedSubstitutionMatrix_bytes / 1e6);
        fprintf(stderr,"threadId = 0: SequenceMaxlength  = %10.1f Mb\n",(float)h_RepeatedSequence_length_vector_bytes / 1e6);
        fprintf(stderr,"threadId = 0: PV                 = %10.1f Mb\n",(float)h_RepeatedPV_bytes / 1e6);
        fprintf(stderr,"threadId = 0: ThickBackbone      = %10.1f Mb\n",(float)h_ThickBackbone_bytes / 1e6);
        fprintf(stderr,"threadId = 0: Overflow           = %10.1f Mb\n",(float)h_overflow_bytes / 1e6);
        fprintf(stderr,"threadId = 0: Total              = %10.1f Mb\n",(double)BytesTransferredToGlobalMemory/1e6);
    }

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Allocating GPU memory...\n");


    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_Database2Dpadded), stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes + 2*num_blocksx * num_threadsx * sizeof(int) ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_RepeatedSubstitutionMatrix), h_RepeatedSubstitutionMatrix_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_RepeatedSequence_length_vector), h_RepeatedSequence_length_vector_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_Hits), stream_blocks[cuda_device_id].h_Hits_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_RepeatedPV), h_RepeatedPV_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_ThickBackbone), h_ThickBackbone_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_overflow), h_overflow_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_RepeatedDiag_array), h_RepeatedDiag_array_bytes ) ;
    gpu_device_calloc( (void**) &(stream_blocks[cuda_device_id].d_GPUBlastInitHitList), stream_blocks[cuda_device_id].h_GPUBlastInitHitList_bytes ) ;

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Done with allocating GPU memory\n");

    // copy host memory to device
    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Copying to GPU memory...\n");



    gpu_memcpy( stream_blocks[cuda_device_id].d_RepeatedSubstitutionMatrix, h_RepeatedSubstitutionMatrix, h_RepeatedSubstitutionMatrix_bytes, cudaMemcpyHostToDevice) ;
    gpu_memcpy( stream_blocks[cuda_device_id].d_RepeatedSequence_length_vector, h_RepeatedSequence_length_vector, h_RepeatedSequence_length_vector_bytes, cudaMemcpyHostToDevice) ;
    gpu_memcpy( stream_blocks[cuda_device_id].d_RepeatedPV, h_RepeatedPV, h_RepeatedPV_bytes, cudaMemcpyHostToDevice) ;
    gpu_memcpy( stream_blocks[cuda_device_id].d_ThickBackbone, h_ThickBackbone, h_ThickBackbone_bytes, cudaMemcpyHostToDevice) ;
    gpu_memcpy( stream_blocks[cuda_device_id].d_overflow, h_overflow, h_overflow_bytes, cudaMemcpyHostToDevice) ;

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Done with copying GPU memory\n");

    // setup execution parameters
    dim3  grid( num_blocksx, num_blocksy, 1);
    dim3  threads( num_threadsx, num_threadsy, 1);

    int  total_shared_mem_size_requested = 0, total_const_mem_size_requested=0;

    total_shared_mem_size_requested = pv_length*sizeof(PV_ARRAY_TYPE) + SUBSTITUTION_MATRIX_LENGTH*sizeof(char) + Sequence_length_vector_stride*sizeof(int) + 3 * num_threadsx * 4 * sizeof(char) + num_threadsx*sizeof(int);

    total_const_mem_size_requested = Query_length + (num_contexts_invoked+1)*sizeof(int)*3;

    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, cuda_device_id);
    if(total_shared_mem_size_requested > deviceProp.sharedMemPerBlock)
    {
        fprintf(stderr,"ERROR: Shared memory requirements (%d bytes per block) bigger than available memory (%d bytes)!!\n",
                total_shared_mem_size_requested, deviceProp.sharedMemPerBlock);
        all_gpu_abnormally_exit(-1);
    }


    if(total_const_mem_size_requested > deviceProp.totalConstMem)
    {
        fprintf(stderr,"ERROR: Constant memory requirements (%d bytes per block) bigger than available memory (%d bytes)!!\n",
                total_const_mem_size_requested, deviceProp.totalConstMem);
        all_gpu_abnormally_exit(-1);
    }

    if((num_threadsx/32)*32 != num_threadsx)
    {
        fprintf(stderr,"ERROR: The no. of threads must be a multipy of 32!\n");
        all_gpu_abnormally_exit(-1);
    }

    int total_subject_length_offset_per_gpu_thread = 0;
    for(int i=0; i<Group_number; i++)
    {
        total_subject_length_offset_per_gpu_thread += Sequence_length_vector[i]+window;
    }

    if(total_subject_length_offset_per_gpu_thread > (1<<15) )// greater than
    {
        fprintf(stderr, "ERROR: the diag. array can not handle this query group!\n");
        all_gpu_abnormally_exit(-1);
    }

    int hash_subject_length = 0;
    for(int i=0; i<Group_number; i++)
    {
        hash_subject_length += Sequence_length_vector[i]+ewp->diag_table->window;
    }

    if(hash_subject_length>(1<<15))
    {
        fprintf(stderr,"ERROR: the  diag. array is not big enough, please make the size of each DB volumn smaller!\n") ;
        all_gpu_abnormally_exit(-1);
    }

    gpu_memcpy_to_symbol( d_Query_const, h_Query, Query_length) ;
    gpu_memcpy_to_symbol( (char*)d_x_dropoff_const, h_x_dropoff, num_contexts_invoked*sizeof(int));
    gpu_memcpy_to_symbol( (char*)d_cutoff_score_const, h_cutoff_score, num_contexts_invoked*sizeof(int));
    gpu_memcpy_to_symbol( (char*)d_query_offset_const, h_query_offset, (num_contexts_invoked+1)*sizeof(int));

    /// For record time elaspsed
    gpu_memcpy( & (stream_blocks[cuda_device_id].d_Database2Dpadded[num_blocksx * num_threadsx]), stream_blocks[cuda_device_id].h_db_access_entry,
                stream_blocks[cuda_device_id].h_GPU_BLASTP_database_bytes,
                cudaMemcpyHostToDevice) ;


    cudaEventRecord(stream_blocks[cuda_device_id].start,stream_blocks[cuda_device_id].stream_id);

    if(NULL == stream_blocks[cuda_device_id].h_db_access_entry)
    {
        printf("Error, the h_db entry is NULL, and the card id is %d\n", cuda_device_id);
        gpu_abnormally_exit(-1);
    }


    // execute the kernel
    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Kernel execution...\n");

    GPU_BLASTP_kernelTwoHit<<< grid, threads,  total_shared_mem_size_requested, stream_blocks[cuda_device_id].stream_id >>>(
        stream_blocks[cuda_device_id].d_Database2Dpadded,
        stream_blocks[cuda_device_id].d_RepeatedSubstitutionMatrix,
        SUBSTITUTION_MATRIX_LENGTH,
        stream_blocks[cuda_device_id].d_RepeatedPV,
        stream_blocks[cuda_device_id].d_ThickBackbone,
        stream_blocks[cuda_device_id].d_overflow,
        stream_blocks[cuda_device_id].d_RepeatedSequence_length_vector,
        Sequence_length_vector_stride,
        num_sequences, pv_length,
        Query_length, diag_mask, diag_array_length,
        num_contexts_invoked,
        lookup->word_length,
        lookup->mask, window,
        window,
        stream_blocks[cuda_device_id].d_RepeatedDiag_array,
        stream_blocks[cuda_device_id].d_Hits,
        stream_blocks[cuda_device_id].d_GPUBlastInitHitList,
        stream_blocks[cuda_device_id].h_GPUBlastInitHitList_offset,
        stream_blocks[cuda_device_id].h_GPUBlastInitHitList_bytes/sizeof(GPUBlastInitHitList) );

    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Done with kernel execution\n");
    cudaEventRecord(stream_blocks[cuda_device_id].event_id, stream_blocks[cuda_device_id].stream_id);

    printf("Finished submitted, and the card id is %d\n", cuda_device_id);

    free(h_x_dropoff);
    free(h_cutoff_score);
    free(h_query_offset);
    free(h_SubstitutionMatrix);

    free(h_RepeatedSequence_length_vector);
    free(h_RepeatedPV);
    free(h_RepeatedSubstitutionMatrix);

}


void memory_defrag(
    const Int4 h_Hits_bytes,
    const Int4 h_GPUBlastInitHitList_bytes,
    const Int4 num_sequences)
{
    int i=0, current_hit_id=0, total_sucessful_extention_for_subject=0;
    GPUBlastInitHitList* GPUBlastInitHitList_tmp=
        (GPUBlastInitHitList*)malloc(h_GPUBlastInitHitList_bytes*sizeof(char));

    GPUBlastInitHitList* current_hit_list_in_this_subject=NULL;

    if((NULL==local_h_Hits) || (NULL==local_h_GPUBlastInitHitList))
    {
        printf("Error: the memory address of local_h_Hits or local_h_GPUBlastInitHitList is unaccessable!\n");
        gpu_abnormally_exit(-1);
    }
    if(NULL==GPUBlastInitHitList_tmp)
    {
        printf("Error: there is no enough memory space to varialbe GPUBlastInitHitList_tmp!\n");
        gpu_abnormally_exit(-1);
    }

    for(i=0; i<num_sequences; i++)
    {
        if( (-1 != local_h_Hits[i*4+3]) && (local_h_Hits[i*4+1]>0)  )
        {
            total_sucessful_extention_for_subject = local_h_Hits[i*4+1];
            current_hit_list_in_this_subject = & (local_h_GPUBlastInitHitList[local_h_Hits[i*4+3]]);
            local_h_Hits[i*4+3] = current_hit_id;

            if( 0==(i&0x10) )
            {
                memcpy(  GPUBlastInitHitList_tmp+current_hit_id,current_hit_list_in_this_subject,total_sucessful_extention_for_subject*sizeof(GPUBlastInitHitList));
            }
            else
            {
                memcpy(  GPUBlastInitHitList_tmp+current_hit_id,  current_hit_list_in_this_subject-total_sucessful_extention_for_subject+1, total_sucessful_extention_for_subject*sizeof(GPUBlastInitHitList));
            }

            current_hit_id += total_sucessful_extention_for_subject;

        }
    }


    gpu_host_mem_free((void**)&local_h_GPUBlastInitHitList);
    local_h_GPUBlastInitHitList = GPUBlastInitHitList_tmp;
    GPUBlastInitHitList_tmp = NULL;

}

void GPU_BLASTP_get_data( Int4* h_Hits,
                          const Int4 h_Hits_bytes,
                          GPUBlastInitHitList* h_GPUBlastInitHitList,
                          const Int4 h_GPUBlastInitHitList_bytes,
                          const Int4 num_sequences)

{


    // copy result from device to host
    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Transferring data from GPU...\n");


    int gpu_device_count=num_of_gpu_cards_used;


    for(int i=0; i<gpu_device_count; i++)
    {
        gpu_stream_synchronize(i);
        ( gpu_memcpy( &(local_h_Hits[stream_blocks[i].h_Hits_offset]), stream_blocks[i].d_Hits,
                      stream_blocks[i].h_Hits_bytes, cudaMemcpyDeviceToHost ) );
        ( gpu_memcpy( &(local_h_GPUBlastInitHitList[stream_blocks[i].h_GPUBlastInitHitList_offset]), stream_blocks[i].d_GPUBlastInitHitList,
                      stream_blocks[i].h_GPUBlastInitHitList_bytes, cudaMemcpyDeviceToHost));
    }

    /// Set the max. gpu wall time
    set_max_gpu_walltime();

    ///memory_defrag
    memory_defrag(h_Hits_bytes, h_GPUBlastInitHitList_bytes, num_sequences);

    memcpy(  h_Hits, local_h_Hits, h_Hits_bytes ) ;
    memcpy( h_GPUBlastInitHitList, local_h_GPUBlastInitHitList, h_GPUBlastInitHitList_bytes ) ;


    if(local_h_Hits!=NULL) gpu_host_mem_free((void**)&local_h_Hits);
    free(local_h_GPUBlastInitHitList);
    local_h_GPUBlastInitHitList=NULL;

    if(h_Database2Dpadded!=NULL) gpu_host_mem_free((void**)&h_Database2Dpadded);


    if( GPU_VERBOSE )
        fprintf(stderr,"threadId = 0: Done transferring data from GPU\n");

    if( GPU_VERBOSE )
    {
        long unsigned int Total_Hits = 0, Total_Ext = 0, Total_Succ_Ext = 0;
        for( int Sequence = 0; Sequence < num_sequences; ++Sequence )
        {
            Total_Hits += h_Hits[ Sequence*H_HITS_SIZE_BYTE + 0 ];
            Total_Succ_Ext += h_Hits[ Sequence*H_HITS_SIZE_BYTE + 1 ];
            Total_Ext += h_Hits[ Sequence*H_HITS_SIZE_BYTE + 2 ];
        }

        fprintf(stderr,"\nTotal_Hits = %lu, Total_Ext = %lu, Total_Succ_Ext = %lu\n",Total_Hits,Total_Ext, Total_Succ_Ext);

    }

}

void GPU_BLASTP_free_memory()
{

    int gpu_device_count=num_of_gpu_cards_used;


    for(int cuda_device_id=0; cuda_device_id<gpu_device_count; cuda_device_id++)
    {
        gpu_set_device(cuda_device_id);

        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_Hits ) ;
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_GPUBlastInitHitList ) ;
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_Database2Dpadded );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_RepeatedSubstitutionMatrix );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_RepeatedSequence_length_vector );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_RepeatedPV );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_ThickBackbone );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_overflow );
        gpu_device_mem_free( (void**)&stream_blocks[cuda_device_id].d_RepeatedDiag_array );
    }

}


Boolean GPU_BLASTP_check_memory(const LookupTableWrap* lookup_wrap,
                                const Blast_ExtendWord* ewp,
                                const BlastGPUOptions* gpu_options,
                                const int h_Database2Dpadded_bytes,
                                const Int4 h_Hits_bytes, const Int4 h_GPUBlastInitHitList_bytes,
                                const Int4 Group_number, const Int4 num_queries,
                                const Int4 query_length
                               )
{

    if( num_queries > NUM_QUERIES_MAX )
        return FALSE;
    if( query_length > QUERY_LENGTH_MAX )
        return FALSE;

    int h_RepeatedSubstitutionMatrix_bytes = SUBSTITUTION_MATRIX_LENGTH * (gpu_options->num_blocksx) * sizeof(char);

    int Sequence_length_vector_stride = ( Group_number / PITCH + ((Group_number % PITCH)?1:0) ) * PITCH;
    long int h_RepeatedSequence_length_vector_bytes = Sequence_length_vector_stride * (gpu_options->num_blocksx) * sizeof(int);

    BlastAaLookupTable *lookup = (BlastAaLookupTable*) lookup_wrap->lut;

    int pv_length = (lookup->backbone_size >> PV_ARRAY_BTS) + 1;
    Round2Multiple(&pv_length);
    int h_RepeatedPV_bytes = (gpu_options->num_blocksx) * pv_length * sizeof(PV_ARRAY_TYPE);

    int h_ThickBackbone_bytes = (lookup->backbone_size +1) * sizeof(AaLookupSmallboneCell);

    int h_overflow_bytes = (lookup->overflow_size+1) * sizeof(Uint2);

    int diag_array_length = ewp->diag_table->diag_array_length;
    unsigned long long int h_RepeatedDiag_array_bytes =
        (unsigned long long int)diag_array_length * (gpu_options->num_blocksx)*(gpu_options->num_threadsx)*sizeof(short);

    if(0==num_of_gpu_cards_used)
    {
        printf("Error: num_of_gpu_cards_used is zero!\n");
        exit(-1);
    }

    int acutal_h_Database2Dpadded_bytes = h_Database2Dpadded_bytes / num_of_gpu_cards_used;
    if(num_of_gpu_cards_used>0)
        acutal_h_Database2Dpadded_bytes *=IMB_factor_of_gpudb_group;

    unsigned long long int Total_global_memory_bytes  =
        acutal_h_Database2Dpadded_bytes + gpu_options->num_blocksx * gpu_options->num_threadsx * sizeof(int) +
        h_RepeatedSubstitutionMatrix_bytes +
        h_RepeatedSequence_length_vector_bytes +
        h_Hits_bytes +
        h_RepeatedPV_bytes +
        h_ThickBackbone_bytes +
        h_overflow_bytes +
        h_RepeatedDiag_array_bytes +
        h_GPUBlastInitHitList_bytes;

    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, 0);


    unsigned long long int GPU_global_memory = (unsigned long long int) (deviceProp.totalGlobalMem*0.93);

    if( Total_global_memory_bytes > GPU_global_memory)
    {
        printf("Not enough mem: %d bytes ,just has %d \n", Total_global_memory_bytes, GPU_global_memory);
        return FALSE;
    }

    return TRUE;
}

Boolean GPU_BLAST_check_availability()
{

    int deviceCount = 0;
    if (cudaGetDeviceCount(&deviceCount) != 0)
        return FALSE;

    if ( 0 == deviceCount )
        return FALSE;

    return TRUE;
}

void GPU_BLASTP_check_rehit_ratio(const Int4* _h_Hits,
                                  const Int4 _num_sequences)
{
    long rehit_count[NO_CELL_LEVELS], re_extension_count[NO_CELL_LEVELS], re_succ_extension_count[NO_CELL_LEVELS],
    total_successful_extension=0, total_hits=0, total_hits_extended=0;

    int icounter, i;

    for(i=0; i<NO_CELL_LEVELS; i++)
    {
        rehit_count[i]=0;
        re_extension_count[i]=0;
        re_succ_extension_count[i]=0;
    }

    for(icounter=0; icounter<_num_sequences; ++icounter)
    {
        total_hits += (_h_Hits)[icounter*4];
        total_hits_extended += (_h_Hits)[icounter*4+2];
        total_successful_extension += (_h_Hits)[icounter*4+1];

        /// Accumulate the re-compute counts for
        /// acutal situation, and the results
        /// are stored in the first position of each
        /// array.
        if((_h_Hits)[icounter*4+3] == -1)
        {
            rehit_count[0] += (_h_Hits)[icounter*4];
            re_extension_count[0] += (_h_Hits)[icounter*4+2];
            re_succ_extension_count[0] +=(_h_Hits)[icounter*4+1];
        }

        for(i=1; i<NO_CELL_LEVELS; i++)
        {
            if((_h_Hits)[icounter*4+1] > (1<<i))
            {
                rehit_count[i] += (_h_Hits)[icounter*4];
                re_extension_count[i] += (_h_Hits)[icounter*4+2];
                re_succ_extension_count[i] +=(_h_Hits)[icounter*4+1];
            }
        }
    }

    printf("total hits: %ld\n",total_hits);
    printf("total_hits_extended: %ld\n", total_hits_extended);
    printf("total successful extensions: %ld\n",total_successful_extension);


    printf("\n============ re-comp. count statistic ================\n");
    printf("cell no. |  re-hits   | re-extended  | re-succ-extened\n");
    printf("--------------------------------------------------------\n");
    printf("autual | %13ld | %13ld | %13ld \n", rehit_count[0],  re_extension_count[0], re_succ_extension_count[0]);
    printf("--------------------------------------------------------\n");
    for(i=1; i<NO_CELL_LEVELS; i++)
    {
        printf("> %3d  | %13ld | %13ld | %13ld \n", (1<<i), rehit_count[i],  re_extension_count[i], re_succ_extension_count[i]);
        printf("--------------------------------------------------------\n");
    }
    printf("========================================================\n");

    if((total_hits>0) && (total_hits_extended>0) && (total_successful_extension>0)  )
    {
        printf("\n=========== re-comp. ratio ( '%' ) statistic =========\n");
        printf("cell no. |   re-hit   | re-extened | re-succ-extended\n");
        printf("--------------------------------------------------------\n");
        printf("autual   | %7lg | %7lg | %7lg\n", 100.0*rehit_count[0]/total_hits,  100.0*re_extension_count[0]/total_hits_extended, 100.0*re_succ_extension_count[0]/total_successful_extension );
        printf("--------------------------------------------------------\n");
        for(i=1; i<NO_CELL_LEVELS; i++)
        {
            printf(" > %3d   | %7lg | %7lg | %7lg\n", (1<<i), 100.0*rehit_count[i]/total_hits,  100.0*re_extension_count[i]/total_hits_extended, 100.0*re_succ_extension_count[i]/total_successful_extension );
            printf("--------------------------------------------------------\n");
        }
        printf("=========================================================\n");
    }
}
